package org.optionalalliance.jrplugin;

import org.optionalalliance.OptionalCheck;

import org.optionalalliance.Optional;

import org.zeroturnaround.bundled.javassist.ClassPool;
import org.zeroturnaround.bundled.javassist.CtClass;
import org.zeroturnaround.bundled.javassist.CtMethod;
import org.zeroturnaround.bundled.javassist.bytecode.LocalVariableAttribute;
import org.zeroturnaround.javarebel.integration.support.JavassistClassBytecodeProcessor;

/**
 * @author Jevgeni Kabanov
 */
public class OptionalCheckCBP extends JavassistClassBytecodeProcessor {
  public void process(ClassPool cp, ClassLoader cl, CtClass ctClass)
      throws Exception {
    if (ctClass.getName().equals("org.optionalalliance.OptionalCheck"))
      return;
    
    boolean optionalCheck = false;
    Object[] anns = ctClass.getAnnotations();
    for (int i = 0; i < anns.length; i++) {
      if (anns[i] instanceof OptionalCheck) {
        optionalCheck = true;
        break;
      }
    }        
    if (!optionalCheck) return;      
    
    CtMethod[] methods = ctClass.getDeclaredMethods();
    for (int methodIndex = 0; methodIndex < methods.length; methodIndex++) {      
      CtMethod ctMethod = methods[methodIndex];
      
      LocalVariableAttribute locVarAttr = 
        ctMethod.getMethodInfo().getCodeAttribute() == null ? null : 
        (LocalVariableAttribute) ctMethod.getMethodInfo().getCodeAttribute().getAttribute(LocalVariableAttribute.tag);

      Object[][] parameterAnnotations = ctMethod.getParameterAnnotations();
      
      for (int paramIndex = 0; paramIndex < parameterAnnotations.length; paramIndex++) {
        if (ctMethod.getParameterTypes()[paramIndex].isPrimitive())
          continue;
        
        Object[] annotations = parameterAnnotations[paramIndex];
        boolean optional = false;
        for (int k = 0; k < annotations.length; k++) {
          if (annotations[k] instanceof Optional) {
            optional = true;
            break;
          }
        }        
        
        if (!optional) {
          String var;
          if (locVarAttr != null)
            var = "'" + locVarAttr.variableName(paramIndex + 1) + "'";
          else 
            var = "" + (paramIndex + 1);
          
          ctMethod.insertBefore(
              "{if ($" + (paramIndex + 1) + " == null) throw new NullPointerException(\"Parameter " + var +" of " + ctMethod.getLongName() + " cannot be null!\");}");
        }
      }
    }
  }
}
